from dataclasses import dataclass
from typing import List, Optional, Union
import torch


@dataclass
class GeoDockInput():
    protein1_embeddings: torch.FloatTensor
    protein2_embeddings: torch.FloatTensor
    pair_embeddings: torch.FloatTensor
    positional_embeddings: torch.FloatTensor


@dataclass
class GeoDockOutput():
    coords: torch.FloatTensor
    rotat: torch.FloatTensor
    trans: torch.FloatTensor
    lddt_logits: torch.FloatTensor
    dist_logits: torch.FloatTensor


import os
import torch
import random
import torch.nn.functional as F
from torch.utils import data
from tqdm import tqdm
from einops import repeat
from geodock.utils.pdb import save_PDB, place_fourth_atom 
from geodock.utils.coords6d import get_coords6d


class GeoDockDataset(data.Dataset):
    def __init__(
        self, 
        dataset: str = 'dips_train_500',
        out_pdb: bool = False,
        out_png: bool = False,
        is_training: bool = True,
        is_testing: bool = False,
        prob: float = 1.0,
        count: int = 0,
        use_Cb: bool = True,
    ):
        self.dataset = dataset 
        self.out_pdb = out_pdb
        self.out_png = out_png
        self.is_training = is_training
        self.is_testing = is_testing
        self.prob = prob
        self.count = count
        self.use_Cb = use_Cb

        if dataset == 'dips_train':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips/pt_files"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips_equidock/train_list_lt_50.txt" 
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        if dataset == 'dips_val':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips/pt_files"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips_equidock/val_list_lt_50.txt" 
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        if dataset == 'dips_train_500':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips/pt_files"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips_equidock/train_list_500.txt" 
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        if dataset == 'dips_val_500':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips/pt_files"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips_equidock/val_list_500.txt" 
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        if dataset == 'dips_test_500':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips/pt_files"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/dips_equidock/test_list_500.txt" 
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 
        
        elif dataset == 'dips_test':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/dips_test"
            self.file_list = [i[:-3] for i in os.listdir(self.data_dir)] 

        elif dataset == 'db5_train_bound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_bound"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/train_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_val_bound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_bound"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/val_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_test_bound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_bound"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/bound_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_train_unbound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_unbound"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/train_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_val_unbound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_unbound"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/val_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_test_flexible_unbound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_unbound_flexible"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/flexible_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 

        elif dataset == 'db5_test_flexible_bound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/db5_bound_flexible"
            self.data_list = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/unbound_list.txt"
            with open(self.data_list, 'r') as f:
                lines = f.readlines()
            self.file_list = [line.strip() for line in lines] 


    def __getitem__(self, idx: int):
        if self.dataset == 'dips_train' or self.dataset == 'dips_val' or self.dataset == 'dips_train_500' or self.dataset == 'dips_val_500' or self.dataset == 'dips_test_500':
            # Get info from file_list 
            _id = self.file_list[idx]
            split_string = _id.split('/')
            _id = split_string[0] + '_' + split_string[1].rsplit('.', 1)[0]
            data = torch.load(os.path.join(self.data_dir, _id+'.pt'))

        else:
            data = torch.load(os.path.join(self.data_dir, self.file_list[idx]+'.pt'))

        # get from data
        _id = data.name
        seq1 = data['receptor'].seq
        coords1 = data['receptor'].pos
        protein1_embeddings = data['receptor'].x
        seq2 = data['ligand'].seq
        coords2 = data['ligand'].pos
        protein2_embeddings = data['ligand'].x

        # crop > 500
        if not self.is_testing:
            crop_size = 500
            if len(seq1) + len(seq2) > crop_size:
                crop_size_per_chain = crop_size // 2
                if len(seq1) > len(seq2):
                    if len(seq2) < crop_size_per_chain:
                        crop_len = crop_size - len(seq2)
                        n = random.randint(0, len(seq1)-crop_len)
                        seq1 = seq1[n:n+crop_len]
                        coords1 = coords1[n:n+crop_len]
                        protein1_embeddings = protein1_embeddings[n:n+crop_len]
                    else:
                        n1 = random.randint(0, len(seq1)-crop_size_per_chain)
                        seq1 = seq1[n1:n1+crop_size_per_chain]
                        coords1 = coords1[n1:n1+crop_size_per_chain]
                        protein1_embeddings = protein1_embeddings[n1:n1+crop_size_per_chain]
                        n2 = random.randint(0, len(seq2)-crop_size_per_chain)
                        seq2 = seq2[n2:n2+crop_size_per_chain]
                        coords2 = coords2[n2:n2+crop_size_per_chain]
                        protein2_embeddings = protein2_embeddings[n2:n2+crop_size_per_chain]

                elif len(seq2) > len(seq1):
                    if len(seq1) < crop_size_per_chain:
                        crop_len = crop_size - len(seq1)
                        n = random.randint(0, len(seq2)-crop_len)
                        seq2 = seq2[n:n+crop_len]
                        coords2 = coords2[n:n+crop_len]
                        protein2_embeddings = protein2_embeddings[n:n+crop_len]
                    else:
                        n1 = random.randint(0, len(seq1)-crop_size_per_chain)
                        seq1 = seq1[n1:n1+crop_size_per_chain]
                        coords1 = coords1[n1:n1+crop_size_per_chain]
                        protein1_embeddings = protein1_embeddings[n1:n1+crop_size_per_chain]
                        n2 = random.randint(0, len(seq2)-crop_size_per_chain)
                        seq2 = seq2[n2:n2+crop_size_per_chain]
                        coords2 = coords2[n2:n2+crop_size_per_chain]
                        protein2_embeddings = protein2_embeddings[n2:n2+crop_size_per_chain]
                else:
                    n = random.randint(0, len(seq1)-crop_size_per_chain)
                    seq1 = seq1[n:n+crop_size_per_chain]
                    coords1 = coords1[n:n+crop_size_per_chain]
                    protein1_embeddings = protein1_embeddings[n:n+crop_size_per_chain]
                    seq2 = seq2[n:n+crop_size_per_chain]
                    coords2 = coords2[n:n+crop_size_per_chain]
                    protein2_embeddings = protein2_embeddings[n:n+crop_size_per_chain]

        try:
            assert len(seq1) == coords1.size(0) == protein1_embeddings.size(0)
            assert len(seq2) == coords2.size(0) == protein2_embeddings.size(0) 
        except:
            print(_id)
        
        # Get ground truth
        label_coords = torch.cat([coords1, coords2], dim=0)
        label_rotat = self.get_rotat(label_coords)
        label_trans = self.get_trans(label_coords)

        # Pair embedding
        input_pairs = self.get_pair_mats(label_coords, len(seq1))
        
        # Contact embedding
        if self.is_training:
            input_contact = self.get_pair_contact(label_coords, len(seq1), count=random.randint(0, 3))
        else:
            input_contact = self.get_pair_contact(label_coords, len(seq1), count=self.count)
            
        pair_embeddings = torch.cat([input_pairs, input_contact], dim=-1)

        # Pair positional embedding
        positional_embeddings = self.get_pair_relpos(len(seq1), len(seq2))

        try:
            assert positional_embeddings.size(0) == pair_embeddings.size(0)
            assert positional_embeddings.size(1) == pair_embeddings.size(1)
        except:
            print(_id)

        if self.out_pdb:
            print(_id)
            test_coords = self.get_full_coords(label_coords)
            out_file = _id + '.pdb'
            if os.path.exists(out_file):
                os.remove(out_file)
                print(f"File '{out_file}' deleted successfully.")
            else:
                print(f"File '{out_file}' does not exist.") 
            save_PDB(out_pdb=out_file, coords=test_coords, seq=seq1+seq2, delim=len(seq1)-1)

        # Output
        output = {
            'id': _id,
            'seq1': seq1,
            'seq2': seq2,
            'protein1_embeddings': protein1_embeddings, 
            'protein2_embeddings': protein2_embeddings, 
            'pair_embeddings': pair_embeddings,
            'positional_embeddings': positional_embeddings,
            'label_rotat': label_rotat,
            'label_trans': label_trans,
            'label_coords': label_coords,
        }
        
        return {key: value for key, value in output.items()}

    def __len__(self):
        return len(self.file_list)
    
    def get_rotat(self, coords):
        # Get backbone coordinates. 
        n_coords = coords[:, 0, :]
        ca_coords = coords[:, 1, :]
        c_coords = coords[:, 2, :]

        # Gram-Schmidt process.
        v1 = c_coords - ca_coords 
        v2 = n_coords - ca_coords
        e1 = F.normalize(v1) 
        u2 = v2 - e1 * (torch.einsum('b i, b i -> b', e1, v2).unsqueeze(-1))
        e2 = F.normalize(u2) 
        e3 = torch.cross(e1, e2, dim=-1)

        # Get rotations.
        rotations=torch.stack([e1, e2, e3], dim=-1)
        return rotations

    def get_trans(self, coords):
        return coords[:, 1, :]

    def get_pair_relpos(self, rec_len, lig_len):
        rmax = 32
        rec = torch.arange(0, rec_len)
        lig = torch.arange(0, lig_len) 
        total = torch.cat([rec, lig], dim=0)
        pairs = total[None, :] - total[:, None]
        pairs = torch.clamp(pairs, min=-rmax, max=rmax)
        pairs = pairs + rmax 
        pairs[:rec_len, rec_len:] = 2*rmax + 1
        pairs[rec_len:, :rec_len] = 2*rmax + 1 
        relpos = F.one_hot(pairs, num_classes=2*rmax+2).float()
        total_len = rec_len + lig_len
        chain_row = torch.cat([torch.zeros(rec_len, total_len), 
                               torch.ones(lig_len, total_len)], dim=0)  
        chain_col = torch.cat([torch.zeros(total_len, rec_len), 
                               torch.ones(total_len, lig_len)], dim=1)
        chains = F.one_hot((chain_row - chain_col + 1).long(), num_classes=3).float()

        pair_pos = torch.cat([relpos, chains], dim=-1)
        return pair_pos
    
    def get_pair_contact(self, coords, n, prob=None, count=None):
        assert (prob is None) ^ (count is None)
        
        if self.use_Cb:
            coords = self.get_full_coords(coords)[:, -1, :]
        else:
            coords = coords[:, 1, :]

        d = torch.norm(coords[:, None, :] - coords[None, :, :], dim=2)
        cutoff = 10.0
        mask = d <= cutoff
        
        mask[:n, :n] = False
        mask[n:, n:] = False

        rec = mask[:n, :n]
        lig = mask[n:, n:]
        inter = mask[:n, n:]

        if prob is not None:        
            random_tensor = torch.rand_like(inter, dtype=torch.float)
            inter = inter & (random_tensor > prob)
        
        elif count is not None:
            # get the indices of all the True values in the tensor
            true_indices = torch.nonzero(inter, as_tuple=False)

            # shuffle the indices
            shuffled_indices = torch.randperm(true_indices.shape[0])

            # make sure count <= # of true indices
            if count > true_indices.shape[0]:
                count = true_indices.shape[0]

            # pick the first shuffled index
            selected_indices = true_indices[shuffled_indices[:count]]

            # create a new tensor of the same shape as the original tensor
            mask = torch.zeros_like(inter)
            
            # set the randomly selected indices to True
            for i in range(selected_indices.shape[0]):
                mask[selected_indices[i, 0], selected_indices[i, 1]] = True

            inter = inter * mask

        upper = torch.cat([rec, inter], dim=1)
        lower = torch.cat([inter.T, lig], dim=1)
        contact = torch.cat([upper, lower], dim=0)
        
        return contact.unsqueeze(-1)

    def get_pair_dist(self, coords, n):
        num_bins = 16
        distogram = self.distogram(
            coords,
            2.0,
            22.0,
            num_bins,
        )

        distogram[:n, n:] = num_bins - 1
        distogram[n:, :n] = num_bins - 1

        # to onehot
        dist = F.one_hot(distogram, num_classes=num_bins).float() 
        
        # test
        if self.out_png:
            data = distogram.numpy()
            plt.imshow(data, cmap='hot', interpolation='nearest')
            plt.colorbar()

            # Save the plot as a PNG file
            plt.savefig('dist.png', dpi=300)
            plt.clf()

        return dist
    
    def get_pair_mats(self, coords, n):
        dist, omega, theta, phi = get_coords6d(coords, use_Cb=self.use_Cb)

        mask = dist < 22.0
        
        num_bins = 16
        dist_bin = self.get_bins(dist, 2.0, 22.0, num_bins)
        omega_bin = self.get_bins(omega, -180.0, 180.0, num_bins)
        theta_bin = self.get_bins(theta, -180.0, 180.0, num_bins)
        phi_bin = self.get_bins(phi, -180.0, 180.0, num_bins)

        def mask_mat(mat):
            mat[~mask] = num_bins - 1
            mat.fill_diagonal_(num_bins - 1)
            mat[:n, n:] = num_bins - 1
            mat[n:, :n] = num_bins - 1
            return mat

        dist_bin[:n, n:] = num_bins - 1
        dist_bin[n:, :n] = num_bins - 1
        omega_bin = mask_mat(omega_bin)
        theta_bin = mask_mat(theta_bin)
        phi_bin = mask_mat(phi_bin)

        # to onehot
        dist = F.one_hot(dist_bin, num_classes=num_bins).float() 
        omega = F.one_hot(omega_bin, num_classes=num_bins).float() 
        theta = F.one_hot(theta_bin, num_classes=num_bins).float() 
        phi = F.one_hot(phi_bin, num_classes=num_bins).float() 
        
        # test
        if self.out_png:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(nrows=2, ncols=2)
            ax[0, 0].imshow(dist_bin.numpy(), cmap='hot', interpolation='nearest')
            ax[0, 1].imshow(omega_bin.numpy(), cmap='hot', interpolation='nearest')
            ax[1, 0].imshow(theta_bin.numpy(), cmap='hot', interpolation='nearest')
            ax[1, 1].imshow(phi_bin.numpy(), cmap='hot', interpolation='nearest')
            
            # Set titles for each plot
            ax[0, 0].set_title('dist')
            ax[0, 1].set_title('omega')
            ax[1, 0].set_title('theta')
            ax[1, 1].set_title('phi')

            # Save the plot as a PNG file
            plt.savefig('dist_orient.png', dpi=300)
            plt.clf()

        return torch.cat([dist, omega, theta, phi], dim=-1)

    def get_full_coords(self, coords):
        #get full coords
        N, CA, C = [x.squeeze(-2) for x in coords.chunk(3, dim=-2)]
        # Infer CB coordinates.
        b = CA - N
        c = C - CA
        a = b.cross(c, dim=-1)
        CB = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + CA
        
        O = place_fourth_atom(torch.roll(N, -1, 0),
                                        CA, C,
                                        torch.tensor(1.231),
                                        torch.tensor(2.108),
                                        torch.tensor(-3.142))
        full_coords = torch.stack(
            [N, CA, C, O, CB], dim=1)
        
        return full_coords

    def distogram(self, coords, min_bin, max_bin, num_bins, use_cb=False):
        # Coords are [... L x 3 x 3], where it's [N, CA, C] x 3 coordinates.
        boundaries = torch.linspace(
            min_bin,
            max_bin,
            num_bins - 1,
            device=coords.device,
        )
        boundaries = boundaries**2
        N, CA, C = [x.squeeze(-2) for x in coords.chunk(3, dim=-2)]

        if use_cb:
            # Infer CB coordinates.
            b = CA - N
            c = C - CA
            a = b.cross(c, dim=-1)
            CB = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + CA
            dists = (CB[..., None, :, :] - CB[..., :, None, :]).pow(2).sum(dim=-1, keepdims=True)
        else:
            dists = (CA[..., None, :, :] - CA[..., :, None, :]).pow(2).sum(dim=-1, keepdims=True)

        bins = torch.sum(dists > boundaries, dim=-1)  # [..., L, L]
        return bins

    def get_bins(self, x, min_bin, max_bin, num_bins):
        # Coords are [... L x 3 x 3], where it's [N, CA, C] x 3 coordinates.
        boundaries = torch.linspace(
            min_bin,
            max_bin,
            num_bins - 1,
            device=x.device,
        )
        bins = torch.sum(x.unsqueeze(-1) > boundaries, dim=-1)  # [..., L, L]
        return bins


if __name__ == '__main__':
    dataset = GeoDockDataset(
        dataset='db5_train_bound',
        out_pdb=True,
        out_png=False,
        is_training=False,
        count=0,
    )

    dataset[0]

    """
    dataloader = data.DataLoader(dataset, batch_size=1, num_workers=6)

    for batch in tqdm(dataloader):
        pass
    """
    



import os
import torch
import torch.nn.functional as F
import numpy as np
from time import time
from geodock.utils.pdb import save_PDB_string, place_fourth_atom 


def dock(
    out_name,
    seq1, 
    seq2,
    model_in,
    model,
    do_refine=True,
    use_openmm=True,
):
    #-----Docking Start-----#
    start_time = time()

    # get prediction
    model_out = model(model_in)

    # coords and plddt
    coords = model_out.coords.squeeze()
    plddt = compute_plddt(model_out.lddt_logits).squeeze()
    coords1, coords2 = coords.split([len(seq1), len(seq2)], dim=0)
    full_coords = torch.cat([get_full_coords(coords1), get_full_coords(coords2)], dim=0)

    # seq
    seq_dict = {'A': seq1, 'B': seq2}
    chains = list(seq_dict.keys())
    delims = np.cumsum([len(s) for s in seq_dict.values()]).tolist()

    # get total len
    total_len = full_coords.size(0)

    # check seq len
    assert len(seq1) + len(seq2) == total_len

    # output dir
    out_dir = './'        
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # get pdb
    out_pdb =  os.path.join(out_dir, f"{out_name}.pdb")

    if os.path.exists(out_pdb):
        os.remove(out_pdb)
        print(f"File '{out_pdb}' deleted successfully.")
    else:
        print(f"File '{out_pdb}' does not exist.") 
        
    pdb_string = save_PDB_string(
        out_pdb=out_pdb, 
        coords=full_coords, 
        error=plddt, 
        seq=seq1+seq2,
        chains=chains,
        delims=delims
    )
    print(f"Completed docking in {time() - start_time:.2f} seconds.")
    #-----Docking end-----#

    #-----Refine start-----#
    if do_refine:
        start_time = time()
        if use_openmm:
            try:
                from geodock.refine.openmm_ref import refine
                refine_input = [out_pdb]
            except:
                exit("OpenMM not installed. Please install OpenMM to use refinement.")
        else:
            try:
                from geodock.refine.pyrosetta_ref import refine
                refine_input = [out_pdb, pdb_string]
            except:
                exit("PyRosetta not installed. Please install PyRosetta to use refinement.")

        refine(*refine_input)
        print(f"Completed refining in {time() - start_time:.2f} seconds.")
    #-----Refine end-----#

def get_full_coords(coords):
    # get full coords
    N, CA, C = [x.squeeze(-2) for x in coords.chunk(3, dim=-2)]
    # Infer CB coordinates.
    b = CA - N
    c = C - CA
    a = b.cross(c, dim=-1)
    CB = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + CA
    
    O = place_fourth_atom(torch.roll(N, -1, 0),
                                    CA, C,
                                    torch.tensor(1.231),
                                    torch.tensor(2.108),
                                    torch.tensor(-3.142))
    full_coords = torch.stack(
        [N, CA, C, O, CB], dim=1)
    
    return full_coords


def compute_plddt(logits: torch.Tensor) -> torch.Tensor:
    num_bins = logits.shape[-1]
    bin_width = 1.0 / num_bins
    bounds = torch.arange(
        start=0.5 * bin_width, end=1.0, step=bin_width, device=logits.device
    )
    probs = F.softmax(logits, dim=-1)
    pred_lddt_ca = torch.sum(
        probs * bounds.view(*((1,) * len(probs.shape[:-1])), *bounds.shape),
        dim=-1,
    )
    return pred_lddt_ca * 100



from typing import Optional
import pytorch_lightning as pl
from torch.utils.data import DataLoader, Dataset, random_split
from geodock.datasets.geodock_dataset import GeoDockDataset


class GeoDockDataModule(pl.LightningDataModule):
    def __init__(
        self,
        train_set: str,
        val_set: str,
        test_set: str,
        batch_size: int = 1,
        is_training: bool = True,
        **kwargs
    ):
        super().__init__()
        self.train_set = train_set
        self.val_set = val_set
        self.test_set = test_set
        self.batch_size = batch_size
        self.is_training = is_training
        self.num_workers = kwargs['num_workers']
        self.pin_memory = kwargs['pin_memory']

        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None
    
    def prepare_data(self):
        pass

    def setup(self, stage: Optional[str] = None):
        self.data_train = GeoDockDataset(
            dataset=self.train_set,
            is_training=self.is_training,
        )
        self.data_val = GeoDockDataset(
            dataset=self.val_set,
            is_training=self.is_training,
        )
        self.data_test = GeoDockDataset(
            dataset=self.test_set,
            is_training=self.is_training,
        )

    def train_dataloader(self):
        return DataLoader(
            dataset=self.data_train,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=True,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.data_val,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )

    def test_dataloader(self):
        return DataLoader(
            dataset=self.data_test,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=False,
        )



###
#   Modified from IPA implementation from https://github.com/lucidrains/invariant-point-attention
###

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops.layers.torch import Rearrange
from einops import rearrange, repeat
from torch import einsum
from geodock.utils.transforms import quaternion_to_matrix


# Helpers
def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d


class InvariantPointAttention(nn.Module):
    def __init__(
        self,
        node_dim,
        edge_dim,
        heads=8,
        scalar_key_dim=16,
        scalar_value_dim=16,
        point_key_dim=4,
        point_value_dim=4,
        eps=1e-8,
    ):
        super().__init__()
        self.eps = eps
        self.heads = heads

        # Number of attentions
        num_attn_logits = 3 

        # QKV projection for scalar attention
        self.scalar_attn_logits_scale = (num_attn_logits * scalar_key_dim) ** -0.5
        self.to_scalar_q = nn.Linear(node_dim, scalar_key_dim * heads, bias=False)
        self.to_scalar_k = nn.Linear(node_dim, scalar_key_dim * heads, bias=False)
        self.to_scalar_v = nn.Linear(node_dim, scalar_value_dim * heads, bias=False)

        # QKV projection for point attention
        point_weight_init_value = torch.log(torch.exp(torch.full((heads,), 1.)) - 1.)
        self.point_weights = nn.Parameter(point_weight_init_value)
        self.point_attn_logits_scale = ((num_attn_logits * point_key_dim) * (9 / 2)) ** -0.5
        self.to_point_q = nn.Linear(node_dim, point_key_dim * heads * 3, bias=False)
        self.to_point_k = nn.Linear(node_dim, point_key_dim * heads * 3, bias=False)
        self.to_point_v = nn.Linear(node_dim, point_value_dim * heads * 3, bias=False)

        # Pairwise representation projection to attention bias
        self.pairwise_attn_logits_scale = num_attn_logits ** -0.5

        self.to_pairwise_attn_bias = nn.Sequential(
            nn.Linear(edge_dim, heads),
            Rearrange('b ... h -> (b h) ...')
        )

        # Combine out - scalar dim + pairwise dim + point dim * (3 for coordinates in R3 and then 1 for norm)
        self.to_out = nn.Linear(heads * (scalar_value_dim + edge_dim + point_value_dim * (3 + 1)), node_dim)

    def forward(
        self,
        node,
        edge,
        mask=None,
        *,
        rotations,
        translations,
    ):
        x, b, h, eps = node, node.size(0), self.heads, self.eps

        # Get queries, keys, values for scalar and point (coordinate-aware) attention pathways
        q_scalar, k_scalar, v_scalar = self.to_scalar_q(x), self.to_scalar_k(x), self.to_scalar_v(x)
        q_point, k_point, v_point = self.to_point_q(x), self.to_point_k(x), self.to_point_v(x)

        # Split out heads
        q_scalar, k_scalar, v_scalar = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h=h), (q_scalar, k_scalar, v_scalar))
        q_point, k_point, v_point = map(lambda t: rearrange(t, 'b n (h d c) -> (b h) n d c', h=h, c=3), (q_point, k_point, v_point))
        rotations = repeat(rotations, 'b n r1 r2 -> (b h) n r1 r2', h=h)
        translations = repeat(translations, 'b n c -> (b h) n () c', h=h)

        # Rotate qkv points into global frame
        q_point = einsum('b n d c, b n r c -> b n d r', q_point, rotations) + translations
        k_point = einsum('b n d c, b n r c -> b n d r', k_point, rotations) + translations
        v_point = einsum('b n d c, b n r c -> b n d r', v_point, rotations) + translations

        # Derive attn logits for scalar and pairwise
        attn_logits_scalar = einsum('b i d, b j d -> b i j', q_scalar, k_scalar) * self.scalar_attn_logits_scale
        
        attn_logits_pairwise = self.to_pairwise_attn_bias(edge) * self.pairwise_attn_logits_scale
        
        # Derive attn logits for point attention
        point_qk_diff = rearrange(q_point, 'b i d c -> b i () d c') - rearrange(k_point, 'b j d c -> b () j d c')
        point_dist = (point_qk_diff ** 2.0).sum(dim=-2)
        point_weights = F.softplus(self.point_weights)
        point_weights = repeat(point_weights, 'h -> (b h) () () ()', b=b)
        attn_logits_points = -0.5 * (point_dist * point_weights * self.point_attn_logits_scale).sum(dim=-1)

        # Add look ahead mask for point attention
        if mask is not None:
            mask = repeat(mask, 'b w l -> (h b) w l', h=h)
            attn_logits_points += (mask * -1e9)

        # Combine attn logits
        attn_logits = attn_logits_scalar + attn_logits_points
        attn_logits = attn_logits + attn_logits_pairwise
        
        # Attention
        attn = attn_logits.softmax(dim=-1)

        # Aggregate values
        results_scalar = einsum('b i j, b j d -> b i d', attn, v_scalar)
        attn_with_heads = rearrange(attn, '(b h) i j -> b h i j', h=h)
        results_pairwise = einsum('b h i j, b i j d -> b h i d', attn_with_heads, edge)

        # Aggregate point values
        results_points = einsum('b i j, b j d c -> b i d c', attn, v_point)

        # Rotate aggregated point values back into local frame
        results_points = einsum('b n d r, b n c r -> b n d c', results_points - translations, rotations.transpose(-1, -2))
        results_points_norm = torch.sqrt(sum(map(torch.square, results_points.unbind(dim=-1))) + eps)

        # Merge back heads
        results_scalar = rearrange(results_scalar, '(b h) n d -> b n (h d)', h=h)
        results_points = rearrange(results_points, '(b h) n d c -> b n (h d c)', h=h)
        results_points_norm = rearrange(results_points_norm, '(b h) n d -> b n (h d)', h=h)

        results = (results_scalar, results_points, results_points_norm)

        results_pairwise = rearrange(results_pairwise, 'b h n d -> b n (h d)', h=h)
        results = (*results, results_pairwise)

        # Concat results and project out
        results = torch.cat(results, dim=-1)
        return self.to_out(results)


class Transition(nn.Module):
    """
    hidden_dim = node_dim
    """
    def __init__(
        self,
        *,
        dim,
    ):
        super().__init__() 
        self.layer1 = nn.Linear(dim, dim)
        self.layer2 = nn.Linear(dim, dim)
        self.layer3 = nn.Linear(dim, dim)
        self.act = nn.ReLU()
    
    def forward(self, x):
        x = self.layer1(x)
        x = self.act(x)
        x = self.layer2(x)
        x = self.act(x)
        x = self.layer3(x)
        return x


class IPABlock(nn.Module):
    """
    dropout = 0.1
    """
    def __init__(
        self,
        *,
        node_dim,
        edge_dim,
        dropout=0.1,
        **kwargs
    ):
        super().__init__()
        self.attn_norm = nn.LayerNorm(node_dim)
        self.attn_dropout = nn.Dropout(dropout)
        self.attn = InvariantPointAttention(node_dim=node_dim, edge_dim=edge_dim, **kwargs)
        
        self.ff_norm = nn.LayerNorm(node_dim)
        self.ff_dropout = nn.Dropout(dropout)
        self.ff = Transition(dim=node_dim)

    def forward(self, node, edge, **kwargs):
        x = self.attn(node, edge, **kwargs) + node
        x = self.attn_dropout(x)
        x = self.attn_norm(x) 

        x = self.ff(x) + x
        x = self.ff_dropout(x)
        x = self.ff_norm(x) 
        return x


class StructureModule(nn.Module):
    def __init__(
        self,
        *,
        node_dim,
        edge_dim,
        depth,
    ):
        super().__init__()
        self.depth = depth
        self.IPA_block = IPABlock(
            node_dim=node_dim, 
            edge_dim=edge_dim
        )

        self.to_update = nn.Linear(node_dim, 6)

    def forward(
        self, 
        node, 
        edge,
        mask=None,
        rotations=None,
        translations=None,
    ):
        device = node.device
        b, n, *_ = node.shape

        # If no initial rotations passed in, start from identity
        if not exists(rotations):
            rotations = torch.eye(3, device=device) # initial rotations
            rotations = repeat(rotations, 'i j -> b n i j', b=b, n=n)

        # If no initial translations passed in, start from identity
        if not exists(translations):
            translations = torch.zeros((b, n, 3), device=device)

        for i in range(self.depth):
            # update nodes and backbone frames
            node = self.IPA_block(
                node,
                edge,
                mask=mask,
                rotations=rotations,
                translations=translations
            )

            quaternions_update, translations_update = self.to_update(node).split([3, 3], dim=-1)
            quaternions_update = F.pad(quaternions_update, (1, 0), value=1.)
            quaternions_update = quaternions_update / torch.linalg.norm(quaternions_update, dim=-1, keepdim=True)
            rotations_update = quaternion_to_matrix(quaternions_update)
            translations = einsum('b n i j, b n j -> b n i', rotations, translations_update) + translations
            rotations = einsum('b n i j, b n j k -> b n i k', rotations, rotations_update)
            
        return node, rotations, translations


# ===test=== 
if __name__ == '__main__':
    node = torch.rand(1, 10, 64)
    edge = torch.rand(1, 10, 10, 64)
    torch.manual_seed(0)
    model = StructureModule(node_dim=64, edge_dim=64, depth=3)
    out = model(node, edge)
    print(out)



import torch


def compute_metrics(model, native):
    # get inputs
    model_rec = model[0].squeeze()
    model_lig = model[1].squeeze()
    native_rec = native[0].squeeze()
    native_lig = native[1].squeeze()
    
    # calc metrics
    c_rmsd = get_c_rmsd(model_rec, model_lig, native_rec, native_lig)
    i_rmsd = get_i_rmsd(model_rec, model_lig, native_rec, native_lig)
    l_rmsd = get_l_rmsd(model_rec, model_lig, native_rec, native_lig)
    fnat = get_fnat(model_rec, model_lig, native_rec, native_lig)
    DockQ = get_DockQ(i_rmsd, l_rmsd, fnat)
    rec_bb_rmsd = get_bb_rmsd(model_rec, native_rec)
    lig_bb_rmsd = get_bb_rmsd(model_lig, native_lig)

    # get interface res
    res1, res2 = get_interface_res(native_rec, native_lig, cutoff=10.0)
    rec_bb_irmsd = get_bb_rmsd(model_rec[res1], native_rec[res1])
    lig_bb_irmsd = get_bb_rmsd(model_lig[res2], native_lig[res2])

    return {'cRMS': c_rmsd, 'iRMS': i_rmsd, 'LRMS': l_rmsd, 'Fnat': fnat, 'DockQ': DockQ, 'Rec_BB_RMS': rec_bb_rmsd, 'Lig_BB_RMS': lig_bb_rmsd, 'Rec_BB_iRMS': rec_bb_irmsd, 'Lig_BB_iRMS': lig_bb_irmsd}
    
def get_interface_res(x1, x2, cutoff=10.0):
    # Calculate pairwise distances
    dist = x1[..., None, :, None, :] - x2[..., None, :, None, :, :]
    dist = (dist ** 2).sum(dim=-1).sqrt().flatten(start_dim=-2)

    # Find minimum distance between each pair of residues
    min_dist, _ = torch.min(dist, dim=-1)

    # Find index < cutoff 
    index = torch.where(min_dist < cutoff)
    res1 = torch.unique(index[0])
    res2 = torch.unique(index[1])
    return res1, res2

def get_c_rmsd(model_rec, model_lig, native_rec, native_lig):
    pred = torch.cat([model_rec, model_lig], dim=0).flatten(end_dim=1)
    label = torch.cat([native_rec, native_lig], dim=0).flatten(end_dim=1)
    R, t = find_rigid_alignment(pred, label)
    pred = (R.mm(pred.T)).T + t
    return get_rmsd(pred, label).item()

def get_i_rmsd(model_rec, model_lig, native_rec, native_lig):
    res1, res2 = get_interface_res(native_rec, native_lig, cutoff=10.0)
    pred = torch.cat([model_rec[res1], model_lig[res2]], dim=0).flatten(end_dim=1)
    label = torch.cat([native_rec[res1], native_lig[res2]], dim=0).flatten(end_dim=1)
    R, t = find_rigid_alignment(pred, label)
    pred = (R.mm(pred.T)).T + t
    return get_rmsd(pred, label).item()

def get_l_rmsd(model_rec, model_lig, native_rec, native_lig):
    model_rec = model_rec.flatten(end_dim=1)
    model_lig = model_lig.flatten(end_dim=1)
    native_rec = native_rec.flatten(end_dim=1)
    native_lig = native_lig.flatten(end_dim=1)
    R, t = find_rigid_alignment(model_rec, native_rec)
    model_lig = (R.mm(model_lig.T)).T + t
    return get_rmsd(model_lig, native_lig).item()

def get_bb_rmsd(model, native):
    pred = model.flatten(end_dim=1)
    label = native.flatten(end_dim=1)
    R, t = find_rigid_alignment(pred, label)
    pred = (R.mm(pred.T)).T + t
    return get_rmsd(pred, label).item()

def get_fnat(model_rec, model_lig, native_rec, native_lig):
    native_res1, native_res2 = get_interface_res(native_rec, native_lig, cutoff=6.0)
    model_res1, model_res2 = get_interface_res(model_rec, model_lig, cutoff=6.0)
    overlap = len(set(native_res1.tolist()) & set(model_res1.tolist())) + len(set(native_res2.tolist()) & set(model_res2.tolist()))
    universe = len(set(native_res1.tolist()) | set(model_res1.tolist())) + len(set(native_res2.tolist()) | set(model_res2.tolist()))
    return float(overlap / (universe + 1e-6))

def get_DockQ(i_rmsd, l_rmsd, fnat):
    i_rmsd_scaled = 1.0 / (1.0 + (i_rmsd/1.5)**2)
    l_rmsd_scaled = 1.0 / (1.0 + (l_rmsd/8.5)**2)
    return (fnat + i_rmsd_scaled + l_rmsd_scaled) / 3

def get_rmsd(pred, label):
    rmsd = torch.sqrt(torch.mean(torch.sum((pred - label) ** 2.0, dim=-1)))
    return rmsd

def find_rigid_alignment(A, B):
    """
    See: https://en.wikipedia.org/wiki/Kabsch_algorithm
    2-D or 3-D registration with known correspondences.
    Registration occurs in the zero centered coordinate system, and then
    must be transported back.
        Args:
        -    A: Torch tensor of shape (N,D) -- Point Cloud to Align (source)
        -    B: Torch tensor of shape (N,D) -- Reference Point Cloud (target)
        Returns:
        -    R: optimal rotation
        -    t: optimal translation
    Test on rotation + translation and on rotation + translation + reflection
    """
    a_mean = A.mean(axis=0)
    b_mean = B.mean(axis=0)
    A_c = A - a_mean
    B_c = B - b_mean
    # Covariance matrix
    H = A_c.T.mm(B_c)
    U, S, Vt = torch.linalg.svd(H)
    # Rotation matrix
    R = Vt.T.mm(U.T)
    
    # special reflection case
    if torch.linalg.det(R) < 0:
        # print("det(R) < R, reflection detected!, correcting for it ...")
        SS = torch.diag(torch.tensor([1.,1.,-1.], device=R.device))
        R = (Vt.T @ SS) @ U.T
    
    # Translation vector
    t = b_mean[None, :] - R.mm(a_mean[None, :].T).T
    t = t.T
    return R, t.squeeze()




import torch
import math
import numpy as np
import scipy
import scipy.spatial
from einops import repeat


def calc_dist(a_coords, b_coords):
    assert a_coords.shape == b_coords.shape
    mat_shape = list(a_coords.shape)
    mat_shape.insert(-1, mat_shape[-2])

    a_coords = a_coords.unsqueeze(-3).expand(mat_shape)
    b_coords = b_coords.unsqueeze(-2).expand(mat_shape)

    dist_mat = (a_coords - b_coords).norm(dim=-1)

    return dist_mat


def calc_dihedral(a_coords,
                  b_coords,
                  c_coords,
                  d_coords,
                  convert_to_degree=True):
    b1 = a_coords - b_coords
    b2 = b_coords - c_coords
    b3 = c_coords - d_coords

    n1 = torch.cross(b1, b2)
    n1 = torch.div(n1, n1.norm(dim=-1, keepdim=True))
    n2 = torch.cross(b2, b3)
    n2 = torch.div(n2, n2.norm(dim=-1, keepdim=True))
    m1 = torch.cross(n1, torch.div(b2, b2.norm(dim=-1, keepdim=True)))

    dihedral = torch.atan2((m1 * n2).sum(-1), (n1 * n2).sum(-1))

    if convert_to_degree:
        dihedral = dihedral * 180 / math.pi

    return dihedral


def calc_planar(a_coords, b_coords, c_coords, convert_to_degree=True):
    v1 = a_coords - b_coords
    v2 = c_coords - b_coords

    a = (v1 * v2).sum(-1)
    b = v1.norm(dim=-1) * v2.norm(dim=-1)

    planar = torch.acos(a / b)

    if convert_to_degree:
        planar = planar * 180 / math.pi

    return planar


# get 6d coordinates from x,y,z coords of N,Ca,C atoms
def get_coords6d(xyz, use_Cb=False):

    n = xyz.shape[0]

    # three anchor atoms
    N  = xyz[..., 0, :]
    Ca = xyz[..., 1, :]
    C  = xyz[..., 2, :]

    # recreate Cb given N,Ca,C
    b = Ca - N
    c = C - Ca
    a = torch.cross(b, c, dim=-1)
    Cb = -0.58273431*a + 0.56802827*b - 0.54067466*c + Ca

    if use_Cb:
        dist = calc_dist(Cb, Cb)
    else:
        dist = calc_dist(Ca, Ca)

    
    omega = calc_dihedral(
                repeat(Ca, 'r i -> r c i', c=n), 
                repeat(Cb, 'r i -> r c i', c=n), 
                repeat(Cb, 'c i -> r c i', r=n), 
                repeat(Ca, 'c i -> r c i', r=n),
            )

    theta = calc_dihedral(
                repeat(N, 'r i -> r c i', c=n), 
                repeat(Ca, 'r i -> r c i', c=n), 
                repeat(Cb, 'r i -> r c i', c=n), 
                repeat(Cb, 'c i -> r c i', r=n),
            )
    phi = calc_planar(
                repeat(Ca, 'r i -> r c i', c=n), 
                repeat(Cb, 'r i -> r c i', c=n), 
                repeat(Cb, 'c i -> r c i', r=n), 
            )


    return dist, omega, theta, phi

if __name__ == '__main__':
    coords = torch.randn(10, 3, 3)
    out = get_coords6d(coords)
    print(out)




import os
import torch
from torch.utils import data
from esm.inverse_folding.util import load_coords


class UnboundDataset(data.Dataset):
    def __init__(
        self,
        dataset: str,
    ):
        self.dataset = dataset
        if self.dataset == 'db5_test_flexible_unbound':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/GeoDock/geodock/benchmark/db5_test_flexible_unbound"
            self.unbound_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/unbound" 
            self.bound_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/bound"
            self.partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/partners"
            self.file_list = [i[:4] for i in os.listdir(self.model_dir) if i[-3:] == 'pdb']

    def __getitem__(self, idx: int):
        if self.dataset == 'db5_test_flexible_unbound':
            _id = self.file_list[idx]
            model_file = os.path.join(self.model_dir, _id+"_p.pdb")
            unbound_file = os.path.join(self.unbound_dir, _id+"_unbound.pdb")
            bound_file = os.path.join(self.bound_dir, _id+"_bound.pdb")
            native_partner_file = os.path.join(self.partner_dir, _id+"_partners") 

            model_partner1 = 'A'
            model_partner2 = 'B'
            with open(native_partner_file, 'r') as f:
                partner = f.readline().strip().split()[1]
                native_partner1 = partner.split('_')[0]
                native_partner2 = partner.split('_')[1]


            model_coords1, model_seq1 = load_coords(model_file, chain=[*model_partner1])
            model_coords2, model_seq2 = load_coords(model_file, chain=[*model_partner2])
            unbound_coords1, unbound_seq1 = load_coords(unbound_file, chain=[*native_partner1])
            unbound_coords2, unbound_seq2 = load_coords(unbound_file, chain=[*native_partner2])
            bound_coords1, bound_seq1 = load_coords(bound_file, chain=[*native_partner1])
            bound_coords2, bound_seq2 = load_coords(bound_file, chain=[*native_partner2])

            model_coords1 = torch.nan_to_num(torch.from_numpy(model_coords1))
            model_coords2 = torch.nan_to_num(torch.from_numpy(model_coords2))
            unbound_coords1 = torch.nan_to_num(torch.from_numpy(unbound_coords1))
            unbound_coords2 = torch.nan_to_num(torch.from_numpy(unbound_coords2))
            bound_coords1 = torch.nan_to_num(torch.from_numpy(bound_coords1))
            bound_coords2 = torch.nan_to_num(torch.from_numpy(bound_coords2))

            # Output
            output = {
                'id': _id[:4],
                'model_coords1': model_coords1, 
                'model_coords2': model_coords2, 
                'unbound_coords1': unbound_coords1,
                'unbound_coords2': unbound_coords2,
                'bound_coords1': bound_coords1,
                'bound_coords2': bound_coords2,
            }
        
        return {key: value for key, value in output.items()}

    def __len__(self):
        return len(self.file_list)
    

if __name__ == '__main__':
    dataset = UnboundDataset(
        dataset='db5_test_flexible_unbound'
    )

    dataset[0]


import os
import torch
from torch.utils import data
from esm.inverse_folding.util import load_coords


class BoundDataset(data.Dataset):
    def __init__(
        self,
        dataset: str,
    ):
        self.dataset = dataset
        if self.dataset.split('_')[0] == 'dips':
            self.native_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/dips_test_random_transformed/complexes"
            self.file_list = [i[:-21] for i in os.listdir(self.native_dir) if i[-3:] == 'pdb'] 
            self.file_list = list(dict.fromkeys(self.file_list)) # remove duplicates
        elif self.dataset.split('_')[0] == 'db5':
            self.native_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/db5_test_random_transformed/complexes"
            self.file_list = [i[:4] for i in os.listdir(self.native_dir) if i[-3:] == 'pdb'] 
            self.file_list = list(dict.fromkeys(self.file_list)) # remove duplicates

        if self.dataset == 'dips_attract':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/dips_attract_results" 
        elif self.dataset == 'dips_cluspro':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/dips_cluspro_results" 
        elif self.dataset == 'dips_patchdock':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/dips_patchdock_results" 
        elif self.dataset == 'dips_equidock':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/dips_equidock_results" 
        elif self.dataset == 'db5_attract':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/db5_attract_results" 
        elif self.dataset == 'db5_cluspro':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/db5_cluspro_results" 
        elif self.dataset == 'db5_patchdock':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/db5_patchdock_results" 
        elif self.dataset == 'db5_equidock':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/repos/equidock_public/test_sets_pdb/db5_equidock_results" 

    def __getitem__(self, idx: int):
        _id = self.file_list[idx] 

        dataset, method = self.dataset.split('_')

        if dataset == 'dips':
            native_rec = os.path.join(self.native_dir, _id+".dill_r_b_COMPLEX.pdb")
            native_lig = os.path.join(self.native_dir, _id+".dill_l_b_COMPLEX.pdb")
            if method == 'attract':
                model_rec = os.path.join(self.model_dir, _id+".dill_r_b_ATTRACT.pdb")
                model_lig = os.path.join(self.model_dir, _id+".dill_l_b_ATTRACT.pdb")
            elif method == 'cluspro':
                model_rec = os.path.join(self.native_dir, _id+".dill_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+".dill_l_b_CLUSPRO.pdb")
            elif method == 'patchdock':
                model_rec = os.path.join(self.native_dir, _id+".dill_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+".dill_l_b_PATCHDOCK.pdb")
            elif method == 'equidock':
                model_rec = os.path.join(self.native_dir, _id+".dill_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+".dill_l_b_EQUIDOCK.pdb")
        elif dataset == 'db5':
            native_rec = os.path.join(self.native_dir, _id+"_r_b_COMPLEX.pdb")
            native_lig = os.path.join(self.native_dir, _id+"_l_b_COMPLEX.pdb")
            if method == 'attract':
                model_rec = os.path.join(self.model_dir, _id+"_r_b_ATTRACT.pdb")
                model_lig = os.path.join(self.model_dir, _id+"_l_b_ATTRACT.pdb")
            elif method == 'cluspro':
                model_rec = os.path.join(self.native_dir, _id+"_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+"_l_b_CLUSPRO.pdb")
            elif method == 'patchdock':
                model_rec = os.path.join(self.native_dir, _id+"_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+"_l_b_PATCHDOCK.pdb")
            elif method == 'equidock':
                model_rec = os.path.join(self.native_dir, _id+"_r_b_COMPLEX.pdb")
                model_lig = os.path.join(self.model_dir, _id+"_l_b_EQUIDOCK.pdb")


        model_coords1, model_seq1 = load_coords(model_rec, chain=None)
        model_coords2, model_seq2 = load_coords(model_lig, chain=None)
        model_coords1 = torch.nan_to_num(torch.from_numpy(model_coords1))
        model_coords2 = torch.nan_to_num(torch.from_numpy(model_coords2))

        native_coords1, native_seq1 = load_coords(native_rec, chain=None)
        native_coords2, native_seq2 = load_coords(native_lig, chain=None)
        native_coords1 = torch.nan_to_num(torch.from_numpy(native_coords1))
        native_coords2 = torch.nan_to_num(torch.from_numpy(native_coords2))

        # Output
        output = {
            'id': _id,
            'model_coords1': model_coords1, 
            'model_coords2': model_coords2, 
            'native_coords1': native_coords1,
            'native_coords2': native_coords2,
        }
        
        return {key: value for key, value in output.items()}

    def __len__(self):
        return len(self.file_list)
    

if __name__ == '__main__':
    dataset = BoundDataset(
        dataset='db5_attract'
    )

    dataset[0]


import esm
import torch
import torch.nn.functional as F
from esm.inverse_folding.util import load_coords
from geodock.utils.coords6d import get_coords6d
from geodock.model.interface import GeoDockInput

def embed(
    seq1,
    seq2,
    coords1,
    coords2,
    model,
    batch_converter,
    device,
):
    assert len(seq1) == coords1.size(0) and len(seq2) == coords2.size(0)

    # Get esm embeddings
    protein1_embeddings = get_esm_rep(seq1, model, batch_converter, device)
    protein2_embeddings = get_esm_rep(seq2, model, batch_converter, device)

    # Get pair embeddings
    coords = torch.cat([coords1, coords2], dim=0)
    input_pairs = get_pair_mats(coords, len(seq1))
    input_contact = torch.zeros(*input_pairs.shape[:-1])[..., None] 
    pair_embeddings = torch.cat([input_pairs, input_contact], dim=-1).to(device)
    
    # Get positional embeddings
    positional_embeddings = get_pair_relpos(len(seq1), len(seq2)).to(device)

    embeddings = GeoDockInput(
        protein1_embeddings=protein1_embeddings.unsqueeze(0),
        protein2_embeddings=protein2_embeddings.unsqueeze(0),
        pair_embeddings=pair_embeddings.unsqueeze(0),
        positional_embeddings=positional_embeddings.unsqueeze(0),
    )

    return embeddings
    
def get_esm_rep(seq_prim, model, batch_converter, device):
    # Use ESM-1b format.
    # The length of tokens is:
    # L (sequence length) + 2 (start and end tokens)
    seq = [
        ("seq", seq_prim)
    ]
    out = batch_converter(seq)
    with torch.no_grad():
        results = model(out[-1].to(device), repr_layers = [33])
        rep = results["representations"][33]
    
    return rep[0, 1:-1, :]

def get_pair_mats(coords, n, use_Cb=True):
    dist, omega, theta, phi = get_coords6d(coords, use_Cb=use_Cb)

    mask = dist < 22.0
    
    num_bins = 16
    dist_bin = get_bins(dist, 2.0, 22.0, num_bins)
    omega_bin = get_bins(omega, -180.0, 180.0, num_bins)
    theta_bin = get_bins(theta, -180.0, 180.0, num_bins)
    phi_bin = get_bins(phi, -180.0, 180.0, num_bins)

    def mask_mat(mat):
        mat[~mask] = num_bins - 1
        mat.fill_diagonal_(num_bins - 1)
        mat[:n, n:] = num_bins - 1
        mat[n:, :n] = num_bins - 1
        return mat

    dist_bin[:n, n:] = num_bins - 1
    dist_bin[n:, :n] = num_bins - 1
    omega_bin = mask_mat(omega_bin)
    theta_bin = mask_mat(theta_bin)
    phi_bin = mask_mat(phi_bin)

    # to onehot
    dist = F.one_hot(dist_bin, num_classes=num_bins).float() 
    omega = F.one_hot(omega_bin, num_classes=num_bins).float() 
    theta = F.one_hot(theta_bin, num_classes=num_bins).float() 
    phi = F.one_hot(phi_bin, num_classes=num_bins).float() 
    
    # test
    return torch.cat([dist, omega, theta, phi], dim=-1)

def get_pair_relpos(rec_len, lig_len):
    rmax = 32
    rec = torch.arange(0, rec_len)
    lig = torch.arange(0, lig_len) 
    total = torch.cat([rec, lig], dim=0)
    pairs = total[None, :] - total[:, None]
    pairs = torch.clamp(pairs, min=-rmax, max=rmax)
    pairs = pairs + rmax 
    pairs[:rec_len, rec_len:] = 2*rmax + 1
    pairs[rec_len:, :rec_len] = 2*rmax + 1 
    relpos = F.one_hot(pairs, num_classes=2*rmax+2).float()
    total_len = rec_len + lig_len
    chain_row = torch.cat([torch.zeros(rec_len, total_len), 
                            torch.ones(lig_len, total_len)], dim=0)  
    chain_col = torch.cat([torch.zeros(total_len, rec_len), 
                            torch.ones(total_len, lig_len)], dim=1)
    chains = F.one_hot((chain_row - chain_col + 1).long(), num_classes=3).float()

    pair_pos = torch.cat([relpos, chains], dim=-1)
    return pair_pos

def get_bins(x, min_bin, max_bin, num_bins):
    # Coords are [... L x 3 x 3], where it's [N, CA, C] x 3 coordinates.
    boundaries = torch.linspace(
        min_bin,
        max_bin,
        num_bins - 1,
        device=x.device,
    )
    bins = torch.sum(x.unsqueeze(-1) > boundaries, dim=-1)  # [..., L, L]
    return bins


import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import repeat, rearrange
from geodock.model.interface import GeoDockOutput

def get_fape(
    coords: torch.Tensor, 
    rotat: torch.Tensor, 
    trans: torch.Tensor,
) -> torch.Tensor:
    """
    """
    size = coords.size(1)
    assert size == trans.size(1) and size == rotat.size(1)
    coords = repeat(coords, 'b c a i -> b r c a i', r=size)
    trans = repeat(trans, 'b r i -> b r c () i', c=size)
    rotat = repeat(rotat, 'b r a i -> b r c a i', c=size)
    fape = torch.einsum('b r c a j, b r c i j -> b r c a i', coords-trans, rotat.transpose(-1, -2))
    return fape

def fape_loss(
    pred: torch.Tensor, 
    label: torch.Tensor, 
    mask: torch.Tensor = None, 
    d_clamp: float = None,
) -> torch.Tensor:
    """
    """
    loss = torch.sqrt(torch.sum((pred - label)**2.0, dim=-1) + 1e-4)
    if d_clamp is not None:
        loss = torch.clamp(loss, max=d_clamp)

    if mask is not None:
        mask = repeat(mask, 'b r c a -> b r c (repeat a)', repeat=loss.size(-1))
        loss = torch.sum(loss * mask) / (torch.sum(mask) + 1e-6)
    else: 
        loss = torch.mean(loss)

    return loss

def inter_fape_loss(
    pred: torch.Tensor, 
    label: torch.Tensor, 
    sep: int, 
    d_clamp: float = None
) -> torch.Tensor:
    """
    """
    loss = torch.sqrt(torch.sum((pred - label)**2.0, dim=-1) + 1e-4)

    if d_clamp is not None:
        loss = torch.clamp(loss, max=d_clamp)

    mask = torch.zeros_like(loss, device=pred.device)

    mask[:, :sep, sep:, :] = 1.0 
    mask[:, sep:, :sep, :] = 1.0

    loss = torch.sum(loss * mask) / torch.sum(mask)

    return loss

def intra_fape_loss(
    pred: torch.Tensor, 
    label: torch.Tensor, 
    sep: int, 
    d_clamp: float = None
) -> torch.Tensor:
    """
    """
    loss = torch.sqrt(torch.sum((pred - label)**2.0, dim=-1) + 1e-4)

    if d_clamp is not None:
        loss = torch.clamp(loss, max=d_clamp)

    mask = torch.zeros_like(loss, device=pred.device)

    mask[:, :sep, :sep, :] = 1.0 
    mask[:, sep:, sep:, :] = 1.0

    loss = torch.sum(loss * mask) / torch.sum(mask)

    return loss

def between_residue_bond_loss(
    pred_coords: torch.Tensor, 
    eps: float = 1e-6, 
    tolerance_factor_soft: float = 12.0
) -> torch.Tensor:
    """
    """
    this_ca_pos = pred_coords[:, :-1, 1]
    this_c_pos = pred_coords[:, :-1, 2]
    next_n_pos = pred_coords[:, 1:, 0]
    next_ca_pos = pred_coords[:, 1:, 1]
    c_n_bond_length = torch.sqrt(
        eps + torch.sum((this_c_pos - next_n_pos) ** 2, dim=-1)
    )
    ca_c_bond_length = torch.sqrt(
        eps + torch.sum((this_ca_pos - this_c_pos) ** 2, dim=-1)
    )
    n_ca_bond_length = torch.sqrt(
        eps + torch.sum((next_n_pos - next_ca_pos) ** 2, dim=-1)
    )
    c_ca_unit_vec = (this_ca_pos - this_c_pos) / ca_c_bond_length[..., None]
    c_n_unit_vec = (next_n_pos - this_c_pos) / c_n_bond_length[..., None]
    n_ca_unit_vec = (next_ca_pos - next_n_pos) / n_ca_bond_length[..., None]

    gt_length = 1.329
    gt_stddev = 0.014
    c_n_bond_length_error = torch.sqrt(eps + (c_n_bond_length - gt_length) ** 2)
    c_n_loss_per_residue = torch.nn.functional.relu(
        c_n_bond_length_error - tolerance_factor_soft * gt_stddev
    )
    c_n_loss = torch.mean(c_n_loss_per_residue)

    ca_c_n_cos_angle = torch.sum(c_ca_unit_vec * c_n_unit_vec, dim=-1)
    gt_angle = -0.4473
    gt_stddev = 0.014
    ca_c_n_cos_angle_error = torch.sqrt(
        eps + (ca_c_n_cos_angle - gt_angle) ** 2
    )
    ca_c_n_loss_per_residue = torch.nn.functional.relu(
        ca_c_n_cos_angle_error - tolerance_factor_soft * gt_stddev
    )
    ca_c_n_loss = torch.mean(ca_c_n_loss_per_residue)
    
    c_n_ca_cos_angle = torch.sum((-c_n_unit_vec) * n_ca_unit_vec, dim=-1)
    gt_angle = -0.5203 
    gt_stddev = 0.03
    c_n_ca_cos_angle_error = torch.sqrt(
        eps + torch.square(c_n_ca_cos_angle - gt_angle)
    )
    c_n_ca_loss_per_residue = torch.nn.functional.relu(
        c_n_ca_cos_angle_error - tolerance_factor_soft * gt_stddev
    )
    c_n_ca_loss = torch.mean(c_n_ca_loss_per_residue)

    loss = c_n_loss + ca_c_n_loss + c_n_ca_loss
    return loss

def violation_loss(
    pred_coords: torch.Tensor, 
    sep: int,
) -> torch.Tensor:
    """
    """
    pred_1 = pred_coords[:, :sep]
    pred_2 = pred_coords[:, sep:]
    loss_1 = between_residue_bond_loss(pred_1) 
    loss_2 = between_residue_bond_loss(pred_2)
    loss = (loss_1 + loss_2) * 0.5
    return loss

def lddt_loss(
    logits: torch.Tensor,
    pred_coords: torch.Tensor,
    label_coords: torch.Tensor,
    cutoff: float = 15.0,
    no_bins: int = 50,
    eps: float = 1e-10,
    **kwargs,
) -> torch.Tensor:
    """
    """
    pred_coords = pred_coords[..., 1, :]
    label_coords = label_coords[..., 1, :]

    score = lddt(
        pred_coords, 
        label_coords, 
        cutoff=cutoff, 
        eps=eps
    )

    score = score.detach()

    bin_index = torch.floor(score * no_bins).long()
    bin_index = torch.clamp(bin_index, max=(no_bins - 1))
    lddt_ca_one_hot = F.one_hot(
        bin_index, num_classes=no_bins
    )

    errors = softmax_cross_entropy(logits, lddt_ca_one_hot)

    # Average over the batch dimension
    loss = torch.mean(errors)

    return loss

def lddt(
    pred_coords: torch.Tensor,
    label_coords: torch.Tensor,
    cutoff: float = 15.0,
    eps: float = 1e-10,
    per_residue: bool = True,
) -> torch.Tensor:
    """
    """
    n = pred_coords.size(1)

    dmat_true = torch.sqrt(
        eps
        + torch.sum(
            (
                label_coords[..., None, :]
                - label_coords[..., None, :, :]
            )
            ** 2,
            dim=-1,
        )
    )

    dmat_pred = torch.sqrt(
        eps
        + torch.sum(
            (
                pred_coords[..., None, :]
                - pred_coords[..., None, :, :]
            )
            ** 2,
            dim=-1,
        )
    )
    dists_to_score = (
        (dmat_true < cutoff)
        * (1.0 - torch.eye(n, device=pred_coords.device))
    )

    dist_l1 = torch.abs(dmat_true - dmat_pred)

    score = (
        (dist_l1 < 0.5).type(dist_l1.dtype)
        + (dist_l1 < 1.0).type(dist_l1.dtype)
        + (dist_l1 < 2.0).type(dist_l1.dtype)
        + (dist_l1 < 4.0).type(dist_l1.dtype)
    )
    score = score * 0.25

    dims = (-1,) if per_residue else (-2, -1)
    norm = 1.0 / (eps + torch.sum(dists_to_score, dim=dims))
    score = norm * (eps + torch.sum(dists_to_score * score, dim=dims))

    return score

def distogram_loss(
    logits,
    coords,
    min_bin=2.3125,
    max_bin=21.6875,
    no_bins=64,
    eps=1e-6,
    **kwargs,
) -> torch.Tensor:
    """
    """
    boundaries = torch.linspace(
        min_bin,
        max_bin,
        no_bins - 1,
        device=logits.device,
    )
    boundaries = boundaries ** 2

    N, CA, C = [x.squeeze(-2) for x in coords.chunk(3, dim=-2)]
    # Infer CB coordinates.
    b = CA - N
    c = C - CA
    a = b.cross(c, dim=-1)
    CB = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + CA

    dists = torch.sum(
        (CB[..., None, :] - CB[..., None, :, :]) ** 2,
        dim=-1,
        keepdims=True,
    )

    true_bins = torch.sum(dists > boundaries, dim=-1)

    errors = softmax_cross_entropy(
        logits,
        F.one_hot(true_bins, no_bins),
    )

    loss = torch.mean(errors)

    return loss

def softmax_cross_entropy(logits, labels):
    loss = -1 * torch.sum(
        labels * F.log_softmax(logits, dim=-1),
        dim=-1,
    )
    return loss

def compute_plddt(logits: torch.Tensor) -> torch.Tensor:
    num_bins = logits.shape[-1]
    bin_width = 1.0 / num_bins
    bounds = torch.arange(
        start=0.5 * bin_width, end=1.0, step=bin_width, device=logits.device
    )
    probs = F.softmax(logits, dim=-1)
    pred_lddt_ca = torch.sum(
        probs * bounds.view(*((1,) * len(probs.shape[:-1])), *bounds.shape),
        dim=-1,
    )
    return pred_lddt_ca * 100


class GeoDockLoss(nn.Module):
    """
    """
    def __init__(self):
        super(GeoDockLoss, self).__init__()

    def forward(self, out: GeoDockOutput, batch):
        pred_coords = out.coords
        pred_rotat = out.rotat
        pred_trans = out.trans
        label_coords = batch['label_coords']
        label_rotat = batch['label_rotat']
        label_trans = batch['label_trans']
        sep = batch['protein1_embeddings'].size(1)
        pred_fape = get_fape(pred_coords, pred_rotat, pred_trans)
        label_fape = get_fape(label_coords, label_rotat, label_trans)

        loss_fns = {
            "intra_loss": lambda: intra_fape_loss(
                pred=pred_fape,
                label=label_fape,
                sep=sep,
                d_clamp=10.0,
            ),
            "inter_loss": lambda: inter_fape_loss(
                pred=pred_fape,
                label=label_fape,
                sep=sep,
                d_clamp=30.0,
            ),
            "dist_loss": lambda: distogram_loss(
                logits=out.dist_logits,
                coords=label_coords,
            ),
            "lddt_loss": lambda: lddt_loss(
                logits=out.lddt_logits,
                pred_coords=pred_coords,
                label_coords=label_coords,
            ),
            "violation_loss": lambda: violation_loss(
                pred_coords=pred_coords,
                sep=sep,
            ),
        }

        losses = {}
        for loss_name, loss_fn in loss_fns.items():
            loss = loss_fn()
            losses[loss_name] = loss
        
        return losses



import os
import torch
from torch.utils import data
from esm.inverse_folding.util import load_coords


class RepD2Dataset(data.Dataset):
    def __init__(
        self,
        dataset: str,
    ):
        self.dataset = dataset
        if self.dataset == 'RepD2':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/top5" 
            self.native_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/native"
            self.model_partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/partners/model"
            self.native_partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/partners/native"
            self.file_list = [i[:6] for i in os.listdir(self.model_dir) if i[-3:] == 'pdb']
            self.file_list = sorted(self.file_list)
        elif self.dataset == 'RepD2_global':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/global/top1" 
            self.native_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/native"
            self.model_partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/partners/model"
            self.native_partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/flexible_set/partners/native"
            self.file_list = [i[:4] for i in os.listdir(self.model_dir) if i[-3:] == 'pdb']
            self.file_list = sorted(self.file_list)
        elif self.dataset == 'RepD2_bound':
            self.model_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/equidock_set/top1" 
            self.native_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/For_Ameya"
            self.model_partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/equidock_set/partners"
            self.file_list = [i[:4] for i in os.listdir(self.model_dir) if i[-3:] == 'pdb']

    def __getitem__(self, idx: int):
        if self.dataset == 'RepD2':
            _id = self.file_list[idx] 
            model_file = os.path.join(self.model_dir, _id+".pdb")
            native_file = os.path.join(self.native_dir, _id[:4]+"_native.pdb")
            model_partner_file = os.path.join(self.model_partner_dir, _id[:4]) 
            native_partner_file = os.path.join(self.native_partner_dir, _id[:4]) 

            with open(model_partner_file, 'r') as f:
                partner = f.readline().strip()
                model_partner1 = partner.split('_')[0]
                model_partner2 = partner.split('_')[1]

            with open(native_partner_file, 'r') as f:
                partner = f.readline().strip()
                native_partner1 = partner.split('_')[0]
                native_partner2 = partner.split('_')[1]

        elif self.dataset == 'RepD2_global':
            _id = self.file_list[idx] 
            model_file = os.path.join(self.model_dir, _id+"_top_model.pdb")
            native_file = os.path.join(self.native_dir, _id[:4]+"_native.pdb")
            model_partner_file = os.path.join(self.model_partner_dir, _id[:4]) 
            native_partner_file = os.path.join(self.native_partner_dir, _id[:4]) 

            with open(model_partner_file, 'r') as f:
                partner = f.readline().strip()
                model_partner1 = partner.split('_')[0]
                model_partner2 = partner.split('_')[1]

            with open(native_partner_file, 'r') as f:
                partner = f.readline().strip()
                native_partner1 = partner.split('_')[0]
                native_partner2 = partner.split('_')[1]
                
            model_coords1, model_seq1 = load_coords(model_file, chain=[*model_partner1])
            model_coords2, model_seq2 = load_coords(model_file, chain=[*model_partner2])
            native_coords1, native_seq1 = load_coords(native_file, chain=[*native_partner1])
            native_coords2, native_seq2 = load_coords(native_file, chain=[*native_partner2])

            model_coords1 = torch.nan_to_num(torch.from_numpy(model_coords1))
            model_coords2 = torch.nan_to_num(torch.from_numpy(model_coords2))
            native_coords1 = torch.nan_to_num(torch.from_numpy(native_coords1))
            native_coords2 = torch.nan_to_num(torch.from_numpy(native_coords2))

        elif self.dataset == 'RepD2_bound':
            _id = self.file_list[idx] 
            model_file = os.path.join(self.model_dir, _id+".pdb")
            native_file_1 = os.path.join(self.native_dir, _id[:4]+"_r_b.pdb")
            native_file_2 = os.path.join(self.native_dir, _id[:4]+"_l_b.pdb")
            model_partner_file = os.path.join(self.model_partner_dir, _id[:4]+"_partners") 

            with open(model_partner_file, 'r') as f:
                partner = f.readline().strip().split()[1]
                model_partner1 = partner.split('_')[0]
                model_partner2 = partner.split('_')[1]

            model_coords1, model_seq1 = load_coords(model_file, chain=[*model_partner1])
            model_coords2, model_seq2 = load_coords(model_file, chain=[*model_partner2])
            native_coords1, native_seq1 = load_coords(native_file_1, chain=[*model_partner1])
            native_coords2, native_seq2 = load_coords(native_file_2, chain=[*model_partner1])

            model_coords1 = torch.nan_to_num(torch.from_numpy(model_coords1))
            model_coords2 = torch.nan_to_num(torch.from_numpy(model_coords2))
            native_coords1 = torch.nan_to_num(torch.from_numpy(native_coords1))
            native_coords2 = torch.nan_to_num(torch.from_numpy(native_coords2))

            # Output
            output = {
                'id': _id[:4],
                'model_coords1': model_coords1, 
                'model_coords2': model_coords2, 
                'native_coords1': native_coords1,
                'native_coords2': native_coords2,
            }
        
        return {key: value for key, value in output.items()}

    def __len__(self):
        return len(self.file_list)
    

if __name__ == '__main__':
    dataset = RepD2Dataset(
        dataset='RepD2_bound'
    )

    for batch in dataset:
        pass



import time
import torch
import torch.nn as nn
import torch.nn.functional as F
import pytorch_lightning as pl
from torch.utils import data
from einops import repeat, rearrange
from geodock.datasets.geodock_dataset import GeoDockDataset
from geodock.model.interface import GeoDockInput, GeoDockOutput
from geodock.model.modules.iterative_transformer import IterativeTransformer
from geodock.utils.loss import GeoDockLoss


class GeoDock(pl.LightningModule):
    def __init__(
        self,
        node_dim: int = 64,
        edge_dim: int = 64,
        gm_depth: int = 1,
        sm_depth: int = 1,
        num_iter: int = 1,
        lr: float = 1e-4,
        weight_decay: float = 0.0,
    ):
        super().__init__()
        # hyperparameters
        self.lr = lr
        self.weight_decay = weight_decay
        self.save_hyperparameters()

        # Embeddings
        esm_dim = 1280
        pair_dim = 65
        positional_dim = 69
        self.esm_to_node = nn.Linear(esm_dim, node_dim)
        self.pair_to_edge = nn.Linear(pair_dim, edge_dim)
        self.positional_to_edge = nn.Linear(positional_dim, edge_dim)

        # Networks
        self.net = IterativeTransformer(
            node_dim=node_dim, 
            edge_dim=edge_dim, 
            gm_depth=gm_depth,
            sm_depth=sm_depth,
            num_iter=num_iter,
        )

        # Loss
        self.loss = GeoDockLoss()

    def forward(self, input: GeoDockInput):
        # Inputs
        protein1_embeddings = input.protein1_embeddings
        protein2_embeddings = input.protein2_embeddings
        pair_embeddings = input.pair_embeddings
        positional_embeddings = input.positional_embeddings

        # Node embedding
        protein_embeddings = torch.cat([protein1_embeddings, protein2_embeddings], dim=1)
        nodes = self.esm_to_node(protein_embeddings)

        # Edge embedding
        edges = self.pair_to_edge(pair_embeddings) + self.positional_to_edge(positional_embeddings)

        # Networks
        lddt_logits, dist_logits, coords, rotat, trans = self.net(
            node=nodes, 
            edge=edges, 
        )

        # Outputs
        output = GeoDockOutput(
            coords=coords,
            rotat=rotat,
            trans=trans,
            lddt_logits=lddt_logits,
            dist_logits=dist_logits,
        )

        return output

    def step(self, batch, batch_idx):
        # Get info from the batch
        protein1_embeddings = batch['protein1_embeddings']
        protein2_embeddings = batch['protein2_embeddings']
        pair_embeddings = batch['pair_embeddings']
        positional_embeddings = batch['positional_embeddings']
        label_rotat = batch['label_rotat']
        label_trans = batch['label_trans']
        label_coords = batch['label_coords']

        # Prepare GeoDock input
        input = GeoDockInput(
            protein1_embeddings=protein1_embeddings,
            protein2_embeddings=protein2_embeddings,
            pair_embeddings=pair_embeddings,
            positional_embeddings=positional_embeddings,
        )

        # Get GeoDock output
        output = self(input)

        # Loss
        losses = self.loss(output, batch)
        intra_loss = losses['intra_loss']
        inter_loss = losses['inter_loss']
        dist_loss = losses['dist_loss']
        lddt_loss = losses['lddt_loss']
        violation_loss = losses['violation_loss']

        if self.current_epoch < 5:
            loss = intra_loss + 0.3*dist_loss + 0.01*lddt_loss 
        elif self.current_epoch < 10:
            loss = intra_loss + inter_loss + 0.3*dist_loss + 0.01*lddt_loss
        else:
            loss = intra_loss + inter_loss + violation_loss + 0.3*dist_loss + 0.01*lddt_loss

        losses.update({'loss': loss})

        return losses

    def _log(self, losses, train=True):
        phase = "train" if train else "val"
        for loss_name, indiv_loss in losses.items():
            self.log(
                f"{phase}/{loss_name}", 
                indiv_loss, 
                on_step=train, on_epoch=(not train), logger=True,
            )

            if(train):
                self.log(
                    f"{phase}/{loss_name}_epoch",
                    indiv_loss,
                    on_step=False, on_epoch=True, logger=True,
                )

    def training_step(self, batch, batch_idx):
        losses = self.step(batch, batch_idx)
        self._log(losses, train=True)
        
        return losses['loss']

    def validation_step(self, batch, batch_idx):
        losses = self.step(batch, batch_idx)
        self._log(losses, train=False)
        
        return losses['loss']

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            filter(lambda p: p.requires_grad, self.parameters()),
            lr=self.lr, 
            weight_decay=self.weight_decay
        )
        return optimizer


if __name__ == '__main__':
    dataset = GeoDockDataset()

    subset_indices = [0]
    subset = data.Subset(dataset, subset_indices)

    #load dataset
    dataloader = data.DataLoader(subset, batch_size=1, num_workers=6)
    
    model = GeoDock()
    trainer = pl.Trainer()
    trainer.validate(model, dataloader)



import os
import esm
import time
import math
import torch
import random
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import torch.nn.functional as F
from tqdm import tqdm
from torch.utils import data
from einops import rearrange, repeat
from esm.inverse_folding.util import load_coords
from src.utils.pdb import save_PDB, place_fourth_atom 
from torch_geometric.data import HeteroData


class GeoDockDataset(data.Dataset):
    def __init__(
        self, 
        save_dir: str,
        dataset: str = 'dips_test',
        out_pdb: bool = False,
        device: str = 'cpu',
    ):
        if dataset == 'dips_test':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/equidock"
            self.file_list = [i[:-21] for i in os.listdir(self.data_dir) if i[-3:] == 'pdb'] 
            self.file_list = list(dict.fromkeys(self.file_list)) # remove duplicates

        elif dataset == 'db5_bound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/structures"
            self.file_list = [i[0:4] for i in os.listdir(self.data_dir) if i[-3:] == 'pdb'] 
            self.file_list = list(dict.fromkeys(self.file_list)) # remove duplicates

        elif dataset == 'db5_unbound':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/db5.5/structures"
            self.file_list = [i[0:4] for i in os.listdir(self.data_dir) if i[-3:] == 'pdb'] 
            self.file_list = list(dict.fromkeys(self.file_list)) # remove duplicates
        
        elif dataset == 'db5_unbound_flexible':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/unbound"
            self.partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/partners"
            self.file_list = [i[:4] for i in os.listdir(self.data_dir) if i[-3:] == 'pdb']

        elif dataset == 'db5_bound_flexible':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/bound"
            self.partner_dir = "/home/lchu11/scr4_jgray21/lchu11/ReplicaDock2/AF_RepD2_set/flexible_targets/partners"
            self.file_list = [i[:4] for i in os.listdir(self.data_dir) if i[-3:] == 'pdb']

        elif dataset == 'abag_test':
            self.data_dir = "/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/exp_aa/cleaned/"
            pdb_list = pd.read_csv("/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/exp_aa/pdb_list.csv")
            self.file_list = pdb_list['pdb_id'].to_list()
            self.partner1 = pdb_list['partner1'].to_list()
            self.partner2 = pdb_list['partner2'].to_list()
        
        self.dataset = dataset
        self.save_dir = save_dir 
        self.out_pdb = out_pdb
        self.device = device

        # Load esm
        esm_model, alphabet = esm.pretrained.load_model_and_alphabet('/home/lchu11/.cache/torch/hub/checkpoints/esm2_t33_650M_UR50D.pt')
        self.batch_converter = alphabet.get_batch_converter()
        self.esm_model = esm_model.to(device).eval()

    def __getitem__(self, idx: int):
        if self.dataset == 'dips_test':
            _id = self.file_list[idx] 
            pdb_file_1 = os.path.join(self.data_dir, _id+".dill_r_b_COMPLEX.pdb")
            pdb_file_2 = os.path.join(self.data_dir, _id+".dill_l_b_COMPLEX.pdb")
            coords1, seq1 = load_coords(pdb_file_1, chain=None)
            coords2, seq2 = load_coords(pdb_file_2, chain=None)
            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))
        
        elif self.dataset == 'db5_bound':
            # Get info from file_list
            _id = self.file_list[idx] 
            pdb_file_1 = os.path.join(self.data_dir, _id+"_r_b.pdb")
            pdb_file_2 = os.path.join(self.data_dir, _id+"_l_b.pdb")
            coords1, seq1 = load_coords(pdb_file_1, chain=None)
            coords2, seq2 = load_coords(pdb_file_2, chain=None)
            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))

        elif self.dataset == 'db5_unbound':
            # Get info from file_list
            _id = self.file_list[idx] 
            pdb_file_1 = os.path.join(self.data_dir, _id+"_r_u.pdb")
            pdb_file_2 = os.path.join(self.data_dir, _id+"_l_u.pdb")
            coords1, seq1 = load_coords(pdb_file_1, chain=None)
            coords2, seq2 = load_coords(pdb_file_2, chain=None)
            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))

        elif self.dataset == 'db5_unbound_flexible':
            _id = self.file_list[idx] 
            print(_id)
            model_file = os.path.join(self.data_dir, _id+"_unbound.pdb")
            partner_file = os.path.join(self.partner_dir, _id+"_partners") 

            with open(partner_file, 'r') as f:
                partner = f.readline().strip().split()[1]
                model_partner1 = partner.split('_')[0]
                model_partner2 = partner.split('_')[1]
                
            coords1, seq1 = load_coords(model_file, chain=[*model_partner1])
            coords2, seq2 = load_coords(model_file, chain=[*model_partner2])

            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))

        elif self.dataset == 'db5_bound_flexible':
            _id = self.file_list[idx] 
            print(_id)
            model_file = os.path.join(self.data_dir, _id+"_bound.pdb")
            partner_file = os.path.join(self.partner_dir, _id+"_partners") 

            with open(partner_file, 'r') as f:
                partner = f.readline().strip().split()[1]
                model_partner1 = partner.split('_')[0]
                model_partner2 = partner.split('_')[1]
                
            coords1, seq1 = load_coords(model_file, chain=[*model_partner1])
            coords2, seq2 = load_coords(model_file, chain=[*model_partner2])

            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))

        elif self.dataset == 'abag_test':
            # Get info from file_list
            _id = self.file_list[idx] 
            partner1 = list(self.partner1[idx])
            partner2 = list(self.partner2[idx])
            pdb_file = os.path.join(self.data_dir, _id+".pdb")
            coords1, seq1 = load_coords(pdb_file, partner1)
            coords2, seq2 = load_coords(pdb_file, partner2)
            coords1 = torch.nan_to_num(torch.from_numpy(coords1))
            coords2 = torch.nan_to_num(torch.from_numpy(coords2))

        # ESM embedding
        esm_rep1 = self.get_esm_rep(seq1)
        esm_rep2 = self.get_esm_rep(seq2)

        if self.out_pdb:
            coords1 = self.get_full_coords(coords1)
            coords2 = self.get_full_coords(coords2)
            test_coords = torch.cat([coords1, coords2], dim=0)
            save_PDB(out_pdb='test.pdb', coords=test_coords, seq=seq1+seq2, delim=len(seq1)-1)

        # save data to a hetero graph 
        data = HeteroData()

        data['receptor'].x = esm_rep1
        data['receptor'].pos = coords1
        data['receptor'].seq = seq1
        data['ligand'].x = esm_rep2
        data['ligand'].pos = coords2
        data['ligand'].seq = seq2
        data.name = _id
        torch.save(data, os.path.join(self.save_dir, _id+'.pt'))

        return coords1

    def __len__(self):
        return len(self.file_list)

    def get_esm_rep(self, seq_prim):
        # Use ESM-1b format.
        # The length of tokens is:
        # L (sequence length) + 2 (start and end tokens)
        seq = [
            ("seq", seq_prim)
        ]
        out = self.batch_converter(seq)
        with torch.no_grad():
            results = self.esm_model(out[-1].to(self.device), repr_layers = [33])
            rep = results["representations"][33].cpu()
        
        return rep[0, 1:-1, :]

    def convert_to_torch_tensor(self, atom_coords):
        # Convert atom_coords to torch tensor.
        n_coords = torch.Tensor(atom_coords['N'])
        ca_coords = torch.Tensor(atom_coords['CA'])
        c_coords = torch.Tensor(atom_coords['C'])
        coords = torch.stack([n_coords, ca_coords, c_coords], dim=1)
        return coords

    def get_full_coords(self, coords):
        #get full coords
        N, CA, C = [x.squeeze(-2) for x in coords.chunk(3, dim=-2)]
        # Infer CB coordinates.
        b = CA - N
        c = C - CA
        a = b.cross(c, dim=-1)
        CB = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + CA
        
        O = place_fourth_atom(torch.roll(N, -1, 0),
                                        CA, C,
                                        torch.tensor(1.231),
                                        torch.tensor(2.108),
                                        torch.tensor(-3.142))
        full_coords = torch.stack(
            [N, CA, C, O, CB], dim=1)
        
        return full_coords


if __name__ == '__main__':
    name = 'db5_unbound_flexible'
    save_dir = '/home/lchu11/scr4_jgray21/lchu11/Docking-dev/data/pts/'+name 
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
        print(f"Created directory: {save_dir}")
    else:
        print(f"Directory {save_dir} already exists")

    dataset = GeoDockDataset(
        save_dir=save_dir,
        dataset=name,
        out_pdb=False,
    )

    dataloader = data.DataLoader(dataset, batch_size=1)

    for batch in tqdm(dataloader):
        pass



###
#   Inspired by FoldingTrunk implementation from https://github.com/facebookresearch/esm/blob/main/esm/esmfold/v1/trunk.py
###

import torch
import torch.nn as nn
import torch.nn.functional as F
import random
from einops import repeat, rearrange
from geodock.model.modules.structure_module import StructureModule
from geodock.model.modules.graph_module import GraphModule 
from geodock.utils.coords6d import get_coords6d
from geodock.utils.transforms import quaternion_to_matrix


class IterativeTransformer(nn.Module):
    def __init__(
        self,
        *,
        node_dim,
        edge_dim,
        gm_depth,
        sm_depth,
        num_iter,
    ):
        super().__init__()
        self.num_iter = num_iter
        self.recycle_bins = 16
        self.graph_module = GraphModule(node_dim=node_dim, edge_dim=edge_dim, depth=gm_depth)
        self.structure_module = StructureModule(node_dim=node_dim, edge_dim=edge_dim, depth=sm_depth)
        self.recycled_node_norm = nn.LayerNorm(node_dim)
        self.recycled_edge_norm = nn.LayerNorm(edge_dim)
        self.recycled_disto = nn.Linear(self.recycle_bins*4, edge_dim)
        self.plddt = PerResidueLDDTCaPredictor(c_in=node_dim, c_hidden=128, no_bins=50)
        self.disto = DistogramHead(c_z=edge_dim, no_bins=64)

    def forward(
        self, 
        node, 
        edge,
        mask=None,
        rotat=None,
        trans=None,
    ):
        device = node.device
        is_grad_enabled = torch.is_grad_enabled()

        # sample the number iterations
        if self.training:
            num_iter = random.randint(1, self.num_iter)
        else:
            num_iter = self.num_iter
        
        # get initial graph features
        node_0 = node
        edge_0 = edge
        recycled_node = torch.zeros_like(node)
        recycled_edge = torch.zeros_like(edge)
        recycled_bins = torch.zeros((*edge.shape[:-1], self.recycle_bins*4), device=device, dtype=torch.float)

        for i in range(num_iter):
            
            is_final_iter = i == (num_iter - 1)
            with torch.set_grad_enabled(is_grad_enabled and is_final_iter):
                # recycling
                recycled_node = self.recycled_node_norm(recycled_node.detach())
                recycled_edge = self.recycled_edge_norm(recycled_edge.detach())
                recycled_edge += self.recycled_disto(recycled_bins.detach())

                # graph module
                node, edge = self.graph_module(node_0 + recycled_node, edge_0 + recycled_edge)

                # structure module
                feats, rotat, trans = self.structure_module(
                    node,
                    edge,
                    mask=mask,
                )

                coords = self.get_coords(rotat, trans)
                
                recycled_node = node
                recycled_edge = edge
        
                # orientogram
                recycled_bins = self.orientogram(
                    coords,
                    self.recycle_bins,
                )

        lddt_logits = self.plddt(feats)
        dist_logits = self.disto(edge)
                
        return lddt_logits, dist_logits, coords, rotat, trans
    
    def orientogram(self, coords, num_bins):
        dist, omega, theta, phi = get_coords6d(coords.squeeze(0), use_Cb=True)

        mask = dist < 22.0
        
        num_bins = 16
        dist_bin = self.get_bins(dist, 2.0, 22.0, num_bins)
        omega_bin = self.get_bins(omega, -180.0, 180.0, num_bins)
        theta_bin = self.get_bins(theta, -180.0, 180.0, num_bins)
        phi_bin = self.get_bins(phi, -180.0, 180.0, num_bins)

        def mask_mat(mat):
            mat[~mask] = num_bins - 1
            mat.fill_diagonal_(num_bins - 1)
            return mat

        omega_bin = mask_mat(omega_bin)
        theta_bin = mask_mat(theta_bin)
        phi_bin = mask_mat(phi_bin)

        # to onehot
        dist = F.one_hot(dist_bin, num_classes=num_bins).float() 
        omega = F.one_hot(omega_bin, num_classes=num_bins).float() 
        theta = F.one_hot(theta_bin, num_classes=num_bins).float() 
        phi = F.one_hot(phi_bin, num_classes=num_bins).float() 
        
        return torch.cat([dist, omega, theta, phi], dim=-1).unsqueeze(0)

    def get_coords(self, rotat, trans):
        # get batch size, total_len
        batch_size, total_len, *_ = rotat.shape

        # initialize residue frames
        input_coords = torch.tensor([[-0.527, 1.359, 0.0], 
                                    [0.0, 0.0, 0.0],
                                    [1.525, 0.0, 0.0]], device=rotat.device)

        input_coords = repeat(input_coords, 'i j -> b n i j', b=batch_size, n=total_len)

        # update coordinates according to predicted rotations and translations   
        coords = torch.einsum('b n a j, b n i j -> b n a i', input_coords, rotat) + trans.unsqueeze(-2)
        return coords

    def get_bins(self, x, min_bin, max_bin, num_bins):
        # Coords are [... L x 3 x 3], where it's [N, CA, C] x 3 coordinates.
        boundaries = torch.linspace(
            min_bin,
            max_bin,
            num_bins - 1,
            device=x.device,
        )
        bins = torch.sum(x.unsqueeze(-1) > boundaries, dim=-1)  # [..., L, L]

        return bins


class PerResidueLDDTCaPredictor(nn.Module):
    def __init__(self, c_in, c_hidden, no_bins):
        super(PerResidueLDDTCaPredictor, self).__init__()

        self.no_bins = no_bins
        self.c_in = c_in
        self.c_hidden = c_hidden
        self.layer_norm = nn.LayerNorm(self.c_in)
        self.linear_1 = nn.Linear(self.c_in, self.c_hidden)
        self.linear_2 = nn.Linear(self.c_hidden, self.c_hidden)
        self.linear_3 = nn.Linear(self.c_hidden, self.no_bins)
        self.relu = nn.ReLU()

    def forward(self, s):
        s = self.layer_norm(s)
        s = self.linear_1(s)
        s = self.relu(s)
        s = self.linear_2(s)
        s = self.relu(s)
        s = self.linear_3(s)

        return s


class DistogramHead(nn.Module):
    """
    Computes a distogram probability distribution.

    For use in computation of distogram loss, subsection 1.9.8
    """

    def __init__(self, c_z, no_bins, **kwargs):
        """
        Args:
            c_z:
                Input channel dimension
            no_bins:
                Number of distogram bins
        """
        super(DistogramHead, self).__init__()

        self.c_z = c_z
        self.no_bins = no_bins
        self.linear = nn.Linear(self.c_z, self.no_bins)

    def forward(self, z):  # [*, N, N, C_z]
        """
        Args:
            z:
                [*, N_res, N_res, C_z] pair embedding
        Returns:
            [*, N, N, no_bins] distogram probability distribution
        """
        # [*, N, N, no_bins]
        logits = self.linear(z)
        logits = logits + logits.transpose(-2, -3)
        
        return logits

# ===test=== 
if __name__ == '__main__':
    node = torch.rand(1, 100, 64)
    edge = torch.rand(1, 100, 100, 32)
    model = IterativeTransformer(node_dim=64, edge_dim=32, gm_depth=3, sm_depth=3, num_iter=2)
    out = model(node=node, edge=edge)
    print(out)



import torch
import numpy as np
from typing import List, Optional, Union


_aa_1_3_dict = {
    'A': 'ALA',
    'C': 'CYS',
    'D': 'ASP',
    'E': 'GLU',
    'F': 'PHE',
    'G': 'GLY',
    'H': 'HIS',
    'I': 'ILE',
    'K': 'LYS',
    'L': 'LEU',
    'M': 'MET',
    'N': 'ASN',
    'P': 'PRO',
    'Q': 'GLN',
    'R': 'ARG',
    'S': 'SER',
    'T': 'THR',
    'V': 'VAL',
    'W': 'TRP',
    'Y': 'TYR',
    '-': 'GAP',
    'X': 'URI',
}


def exists(x):
    return x is not None


def place_fourth_atom(a_coord: torch.Tensor, b_coord: torch.Tensor,
                      c_coord: torch.Tensor, length: torch.Tensor,
                      planar: torch.Tensor,
                      dihedral: torch.Tensor) -> torch.Tensor:
    """
    Given 3 coords + a length + a planar angle + a dihedral angle, compute a fourth coord
    """
    bc_vec = b_coord - c_coord
    bc_vec = bc_vec / bc_vec.norm(dim=-1, keepdim=True)

    n_vec = (b_coord - a_coord).expand(bc_vec.shape).cross(bc_vec)
    n_vec = n_vec / n_vec.norm(dim=-1, keepdim=True)

    m_vec = [bc_vec, n_vec.cross(bc_vec), n_vec]
    d_vec = [
        length * torch.cos(planar),
        length * torch.sin(planar) * torch.cos(dihedral),
        -length * torch.sin(planar) * torch.sin(dihedral)
    ]

    d_coord = c_coord + sum([m * d for m, d in zip(m_vec, d_vec)])
    return d_coord


def save_PDB(out_pdb: str,
             coords: torch.Tensor,
             seq: str,
             b_factors: torch.Tensor = None,
             delim: int = None) -> None:
    """
    Write set of N, CA, C, O, CB coords to PDB file
    """

    if type(delim) == type(None):
        delim = -1
    
    if b_factors is None:
        b_factors = torch.zeros(coords.size(0), device=coords.device)

    atoms = ['N', 'CA', 'C', 'O', 'CB']

    with open(out_pdb, "a") as f:
        k = 0
        for r, residue in enumerate(coords):
            AA = _aa_1_3_dict[seq[r]]
            for a, atom in enumerate(residue):
                if AA == "GLY" and atoms[a] == "CB": continue
                x, y, z = atom
                f.write(
                    "ATOM  %5d  %-2s  %3s %s%4d    %8.3f%8.3f%8.3f  %4.2f %4.2f\n"
                    % (k + 1, atoms[a], AA, "A" if r <= delim else "B", r + 1,
                       x, y, z, 1, b_factors[r]))
                k += 1
        f.close()


def save_PDB_string(
    out_pdb: str,
    coords: torch.Tensor,
    seq: str,
    chains: List[str] = None,
    error: torch.Tensor = None,
    delims: Union[int, List[int]] = None,
    atoms=['N', 'CA', 'C', 'O', 'CB'],
    write_pdb=True,
) -> None:
    """
    Write set of N, CA, C, O, CB coords to PDB file
    """

    if not exists(chains):
        chains = ["A", "B"]

    if type(delims) == type(None):
        delims = -1
    elif type(delims) == int:
        delims = [delims]

    if not exists(error):
        error = torch.zeros(len(seq))

    pdb_string = ""
    k = 0
    for r, residue in enumerate(coords):
        AA = _aa_1_3_dict[seq[r]]
        for a, atom in enumerate(residue):
            if AA == "GLY" and atoms[a] == "CB": continue
            x, y, z = atom
            chain_id = chains[np.where(np.array(delims) - r > 0)[0][0]]
            pdb_string += "ATOM  %5d  %-2s  %3s %s%4d    %8.3f%8.3f%8.3f  %4.2f  %4.2f\n" % (
                k + 1, atoms[a], AA, chain_id, r + 1, x, y, z, 1, error[r])
            k += 1

            if k in delims:
                pdb_string += "TER  %5d      %3s %s%4d\n" % (
                    k + 1, AA, chain_id, r + 1)
                
    pdb_string += "END\n"

    if write_pdb:
        with open(out_pdb, "w") as f:
            f.write(pdb_string)

    return pdb_string


import hydra
import torch
import os
from omegaconf import DictConfig

#os.environ["WANDB_API_KEY"] = 'bb76c909bdc55c9510c7e47d567c2f6c8a30f369'
#os.environ["WANDB_MODE"] = "offline"


@hydra.main(config_path="config/", config_name="config.yaml")
def main(config: DictConfig):
    torch.manual_seed(0)

    # Imports can be nested inside @hydra.main to optimize tab completion
    # https://github.com/facebookresearch/hydra/issues/934
    from geodock.trainer.train import train
    from geodock.trainer import utils

    # A couple of optional utilities:
    # - disabling python warnings
    # - forcing debug-friendly configuration
    # - verifying experiment name is set when running in experiment mode
    # You can safely get rid of this line if you don't want those
    utils.extras(config)

    # Pretty print config using Rich library
    if config.get("print_config"):
        utils.print_config(config, resolve=True)

    # Train model
    return train(config)


if __name__ == "__main__":
    main()
